#!/usr/bin/python
# SPDX-License-Identifier: ISC
#
# Copyright (c) 2020 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
Following tests are covered to test bgp recursive route and ebgp
multi-hop functionality:

1.  Verify that BGP routes are installed in iBGP peer, only when there
    is a recursive route for next-hop reachability.
2.  Verify that any BGP prefix received with next hop as self-ip is
    not installed in BGP RIB or FIB table.
3.  Verify password authentication for eBGP and iBGP peers.
4.  Verify that for a BGP prefix next-hop information doesn't change
    when same prefix is received from another peer via recursive lookup.
5.  Verify that BGP path attributes are present in CLI outputs and
    JSON format, even if set to default.
6.  Verifying the BGP peering between loopback and physical link's IP
    of 2 peer routers.
7.  Verify that BGP Active/Standby/Pre-emption/ECMP.
"""

import os
import sys
import time
import pytest
from time import sleep

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

# Import topoJson from lib, to create topology and initial configuration
from lib.common_config import (
    start_topology,
    write_test_header,
    apply_raw_config,
    write_test_footer,
    reset_config_on_routers,
    verify_rib,
    create_static_routes,
    check_address_types,
    step,
    create_route_maps,
    create_interface_in_kernel,
    shutdown_bringup_interface,
)
from lib.topolog import logger
from lib.bgp import (
    verify_bgp_convergence,
    create_router_bgp,
    verify_bgp_rib,
    verify_bgp_convergence_from_running_config,
    modify_as_number,
    verify_bgp_attributes,
    clear_bgp,
)
from lib.topojson import build_config_from_json


pytestmark = [pytest.mark.bgpd, pytest.mark.staticd]


# Global variables
BGP_CONVERGENCE = False
KEEP_ALIVE_TIMER = 2
HOLD_DOWN_TIMER = 6
ADDR_TYPES = check_address_types()
NETWORK = {
    "ipv4": ["100.1.1.1/32", "100.1.1.2/32"],
    "ipv6": ["100::1/128", "100::2/128"],
}

RECUR_NEXT_HOP = {
    "N1": {"ipv4": "20.20.20.20/24", "ipv6": "20:20::20:20/120"},
    "N2": {"ipv4": "30.30.30.30/24", "ipv6": "30:30::30:30/120"},
    "N3": {"ipv4": "40.40.40.40/24", "ipv6": "40:40::40:40/120"},
}

CHANGED_NEXT_HOP = {
    "4thOctate": {"ipv4": "10.0.1.250/24", "ipv6": "fd00:0:0:1::100/64"},
    "3rdOctate": {"ipv4": "10.0.10.2/24", "ipv6": "fd00:0:0:10::2/64"},
}

Loopabck_IP = {
    "Lo_R1": {"ipv4": "1.1.1.1/32", "ipv6": "1:1::1:1/128"},
    "Lo_R4": {"ipv4": "4.4.4.4/32", "ipv6": "4:4::4:4/128"},
}


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_recursive_route_ebgp_multi_hop.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    global BGP_CONVERGENCE
    BGP_CONVERGENCE = verify_bgp_convergence(tgen, topo)
    assert BGP_CONVERGENCE is True, "setup_module : Failed \n Error : {}".format(
        BGP_CONVERGENCE
    )

    logger.info("Running setup_module() done")


def teardown_module():
    """Teardown the pytest environment"""

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
#
#   Tests starting
#
#####################################################


def test_recursive_routes_iBGP_peer_p1(request):
    """
    Verify that BGP routes are installed in iBGP peer, only
    when there is a recursive route for next-hop reachability.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config :Configure BGP neighborship between R1 and R3.")
    reset_config_on_routers(tgen)

    dut = "r1"
    protocol = "static"

    step(
        "Configure static routes on R1 pointing next-hop as connected"
        "link between R1 & R3's IP"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": topo["routers"]["r3"]["links"]["r1"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_4)

        step(
            "Verify on router R1 that these static routes are "
            "installed in RIB+FIB of R1"
        )
        result = verify_rib(
            tgen,
            addr_type,
            dut,
            input_dict_4,
            next_hop=topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0],
            protocol=protocol,
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Redistribute these static routes in BGP on router R1")
    input_dict_2 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                    "ipv6": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_2)

    step(
        "Verify on router R1 that these static routes are installed"
        "in RIB table as well"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": topo["routers"]["r3"]["links"]["r1"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = verify_bgp_rib(
            tgen,
            addr_type,
            dut,
            input_dict_4,
            next_hop=topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0],
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    step(
        "Configure a static routes for next hop IP on R2 via multiple"
        "recursive static routes"
    )
    dut = "r2"
    create_interface_in_kernel(
        tgen, dut, "lo10", "40.40.40.50", netmask="255.255.255.0", create=True
    )
    create_interface_in_kernel(
        tgen, dut, "lo10", "40:40::40:50", netmask="120", create=True
    )
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
            "r2": {
                "static_routes": [
                    {
                        "network": topo["routers"]["r3"]["links"]["r1"][addr_type],
                        "next_hop": RECUR_NEXT_HOP["N1"][addr_type].split("/")[0],
                    },
                    {
                        "network": RECUR_NEXT_HOP["N1"][addr_type],
                        "next_hop": RECUR_NEXT_HOP["N2"][addr_type].split("/")[0],
                    },
                    {
                        "network": RECUR_NEXT_HOP["N2"][addr_type],
                        "next_hop": RECUR_NEXT_HOP["N3"][addr_type].split("/")[0],
                    },
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_3)
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

        step("verify if redistributed routes are now installed in FIB of R2")
        result = verify_rib(
            tgen,
            addr_type,
            "r2",
            input_dict_4,
            next_hop=topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0],
            protocol="bgp",
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    step("Delete 1 route from static recursive for the next-hop IP")
    dut = "r2"
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
            "r2": {
                "static_routes": [
                    {
                        "network": RECUR_NEXT_HOP["N1"][addr_type],
                        "next_hop": RECUR_NEXT_HOP["N2"][addr_type].split("/")[0],
                        "delete": True,
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_3)
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

        step("Verify that redistributed routes are withdrawn from FIB of R2")
        result = verify_rib(
            tgen,
            addr_type,
            dut,
            input_dict_4,
            next_hop=topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0],
            protocol="bgp",
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Routes are still present \n Error: {}".format(
            tc_name, result
        )

    step("Reconfigure the same static route on R2 again")
    dut = "r2"
    for addr_type in ADDR_TYPES:
        input_dict_3 = {
            "r2": {
                "static_routes": [
                    {
                        "network": RECUR_NEXT_HOP["N1"][addr_type],
                        "next_hop": RECUR_NEXT_HOP["N2"][addr_type].split("/")[0],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_3)
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

        step("Verify that redistributed routes are again installed" "in FIB of R2")
        result = verify_rib(
            tgen,
            addr_type,
            dut,
            input_dict_4,
            next_hop=topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0],
            protocol="bgp",
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    step("Configure static route with changed next-hop from same subnet")
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": topo["routers"]["r3"]["links"]["r1"][
                            addr_type
                        ].split("/")[0],
                        "delete": True,
                    },
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": CHANGED_NEXT_HOP["4thOctate"][addr_type].split("/")[
                            0
                        ],
                    },
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_4)
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

        result = verify_rib(tgen, addr_type, "r1", input_dict_4, protocol="static")
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

        step(
            "Verify that redistributed routes are not withdrawn as changed"
            "next-hop IP, belongs to the same subnet"
        )
        result = verify_rib(tgen, addr_type, "r2", input_dict_4, protocol="bgp")
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Configure static route with changed next-hop from different subnet")
    dut = "r1"
    create_interface_in_kernel(
        tgen, dut, "lo10", "10.0.10.10", netmask="255.255.255.0", create=True
    )
    create_interface_in_kernel(
        tgen, dut, "lo10", "fd00:0:0:10::104", netmask="64", create=True
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": CHANGED_NEXT_HOP["4thOctate"][addr_type].split("/")[
                            0
                        ],
                        "delete": True,
                    },
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": CHANGED_NEXT_HOP["3rdOctate"][addr_type].split("/")[
                            0
                        ],
                    },
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_4)
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

        result = verify_rib(tgen, addr_type, "r1", input_dict_4, protocol="static")
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

        step(
            "Verify that redistributed routes are withdrawn as changed "
            "next-hop IP, belongs to different subnet"
        )
        result = verify_rib(
            tgen, addr_type, "r2", input_dict_4, protocol="bgp", expected=False
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Routes are still present \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_next_hop_as_self_ip_p1(request):
    """
    Verify that any BGP prefix received with next hop as
    self-ip is not installed in BGP RIB or FIB table.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config :Configure BGP neighborship between R1 and R3.")
    reset_config_on_routers(tgen)

    step(
        "Configure static routes on R1 with a next-hop IP belonging"
        "to the same subnet of R2's link IP."
    )
    dut = "r1"
    create_interface_in_kernel(
        tgen,
        dut,
        "lo10",
        topo["routers"]["r4"]["links"]["r2"]["ipv4"].split("/")[0],
        netmask="255.255.255.0",
        create=True,
    )
    create_interface_in_kernel(
        tgen,
        dut,
        "lo10",
        topo["routers"]["r4"]["links"]["r2"]["ipv6"].split("/")[0],
        netmask="64",
        create=True,
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": topo["routers"]["r2"]["links"]["r4"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_4)

        step("Verify that static routes are installed in RIB and FIB of R1")
        result = verify_rib(
            tgen,
            addr_type,
            dut,
            input_dict_4,
            next_hop=topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
            protocol="static",
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Redistribute static routes into BGP on R1")
    input_dict_2 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                    "ipv6": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_2)

    step(
        "Verify that R2 denies the prefixes received in update message,"
        "as next-hop IP belongs to connected interface"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": topo["routers"]["r2"]["links"]["r4"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = verify_bgp_rib(
            tgen,
            addr_type,
            "r2",
            input_dict_4,
            next_hop=topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Routes are still present \n Error: {}".format(
            tc_name, result
        )

    step("Shut interface on R2 that has IP from the subnet as BGP next-hop")
    intf_r2_r4 = topo["routers"]["r2"]["links"]["r4"]["interface"]
    shutdown_bringup_interface(tgen, "r2", intf_r2_r4)

    for addr_type in ADDR_TYPES:
        clear_bgp(tgen, addr_type, "r2")
    step(
        "Verify that redistributed routes now appear only in BGP table,"
        "as next-hop IP is no more active on R2"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": topo["routers"]["r2"]["links"]["r4"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = verify_bgp_rib(
            tgen,
            addr_type,
            "r2",
            input_dict_4,
            next_hop=topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    step("No shutdown interface on R2 which was shut in previous step")
    intf_r2_r4 = topo["routers"]["r2"]["links"]["r4"]["interface"]
    shutdown_bringup_interface(tgen, "r2", intf_r2_r4, ifaceaction=True)

    step(
        "Verify that R2 dosn't install prefixes RIB to FIB as next-hop"
        "interface is up now"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": NETWORK[addr_type],
                        "next_hop": topo["routers"]["r2"]["links"]["r4"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = verify_bgp_rib(
            tgen,
            addr_type,
            "r2",
            input_dict_4,
            next_hop=topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )
        result = verify_rib(
            tgen,
            addr_type,
            "r2",
            input_dict_4,
            next_hop=topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Routes are still present \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_next_hop_with_recursive_lookup_p1(request):
    """
    Verify that for a BGP prefix next-hop information doesn't change
    when same prefix is received from another peer via recursive lookup.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config :Configure BGP neighborship between R1 and R3.")
    reset_config_on_routers(tgen)

    step("Verify that BGP peering comes up.")

    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Do redistribute connected on router R3.")
    input_dict_1 = {
        "r3": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {"redistribute": [{"redist_type": "connected"}]}
                    },
                    "ipv6": {
                        "unicast": {"redistribute": [{"redist_type": "connected"}]}
                    },
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify that R1 receives all connected")
    for addr_type in ADDR_TYPES:
        routes = {
            "ipv4": ["1.0.3.17/32", "10.0.1.0/24", "10.0.3.0/24"],
            "ipv6": ["2001:db8:f::3:17/128", "fd00:0:0:1::/64", "fd00:0:0:3::/64"],
        }
        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        result = verify_rib(tgen, addr_type, "r1", input_dict, protocol="bgp")
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step(
        "Configure a BGP neighborship between R1 and R4, directly via "
        "eBGP multi-hop."
    )
    r1_local_as = topo["routers"]["r1"]["bgp"]["local_as"]
    r1_r3_addr = topo["routers"]["r1"]["links"]["r3"]
    r4_local_as = topo["routers"]["r4"]["bgp"]["local_as"]
    r4_r3_addr = topo["routers"]["r4"]["links"]["r3"]
    ebgp_multi_hop = 3

    for addr_type in ADDR_TYPES:
        raw_config = {
            "r1": {
                "raw_config": [
                    "router bgp {}".format(r1_local_as),
                    "neighbor {} remote-as {}".format(
                        r4_r3_addr[addr_type].split("/")[0], r4_local_as
                    ),
                    "neighbor {} timers {} {}".format(
                        r4_r3_addr[addr_type].split("/")[0],
                        KEEP_ALIVE_TIMER,
                        HOLD_DOWN_TIMER,
                    ),
                    "neighbor {} ebgp-multihop {}".format(
                        r4_r3_addr[addr_type].split("/")[0], ebgp_multi_hop
                    ),
                ]
            },
            "r4": {
                "raw_config": [
                    "router bgp {}".format(r4_local_as),
                    "neighbor {} remote-as {}".format(
                        r1_r3_addr[addr_type].split("/")[0], r1_local_as
                    ),
                    "neighbor {} timers {} {}".format(
                        r1_r3_addr[addr_type].split("/")[0],
                        KEEP_ALIVE_TIMER,
                        HOLD_DOWN_TIMER,
                    ),
                    "neighbor {} ebgp-multihop {}".format(
                        r1_r3_addr[addr_type].split("/")[0], ebgp_multi_hop
                    ),
                ]
            },
        }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    for addr_type in ADDR_TYPES:
        if addr_type == "ipv4":
            raw_config = {
                "r1": {
                    "raw_config": [
                        "router bgp {}".format(r1_local_as),
                        "address-family {} unicast".format(addr_type),
                        "no neighbor {} activate".format(
                            r4_r3_addr["ipv6"].split("/")[0]
                        ),
                    ]
                },
                "r4": {
                    "raw_config": [
                        "router bgp {}".format(r4_local_as),
                        "address-family {} unicast".format(addr_type),
                        "no neighbor {} activate".format(
                            r1_r3_addr["ipv6"].split("/")[0]
                        ),
                    ]
                },
            }
        else:
            raw_config = {
                "r1": {
                    "raw_config": [
                        "router bgp {}".format(r1_local_as),
                        "address-family {} unicast".format(addr_type),
                        "neighbor {} activate".format(
                            r4_r3_addr[addr_type].split("/")[0]
                        ),
                    ]
                },
                "r4": {
                    "raw_config": [
                        "router bgp {}".format(r4_local_as),
                        "address-family {} unicast".format(addr_type),
                        "neighbor {} activate".format(
                            r1_r3_addr[addr_type].split("/")[0]
                        ),
                    ]
                },
            }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    step("Verify that BGP session between R1 and R4 comes up" "(recursively via R3).")
    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Do redistribute connected on router R4.")
    input_dict_1 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {"redistribute": [{"redist_type": "connected"}]}
                    },
                    "ipv6": {
                        "unicast": {"redistribute": [{"redist_type": "connected"}]}
                    },
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step(
        "Verify that R1 now receives BGP prefix of link r3-r4 via 2 "
        "next-hops R3 and R4. however do not install with NHT R4 in FIB."
    )
    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]

        result = verify_rib(
            tgen, addr_type, "r1", input_dict, protocol="bgp", next_hop=next_hop
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Clear bgp sessions from R1 using 'clear ip bgp *'")
    for addr_type in ADDR_TYPES:
        clear_bgp(tgen, addr_type, "r1")

    step(
        "Verify that prefix of link r3-r4 is again learned via 2 "
        "next-hops (from R3 and R4 directly)"
    )
    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]

        result = verify_rib(
            tgen, addr_type, "r1", input_dict, protocol="bgp", next_hop=next_hop
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Remove redistribution from router R3.")
    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": "300",
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "redistribute": [
                                {"redist_type": "connected", "delete": True}
                            ]
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "redistribute": [
                                {"redist_type": "connected", "delete": True}
                            ]
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step(
        "Verify that peering between R1-R4 goes down and prefix "
        "of link r3-r4, with NHT R4 is withdrawn."
    )

    logger.info("Sleeping for holddowntimer: {}".format(HOLD_DOWN_TIMER))
    sleep(HOLD_DOWN_TIMER + 1)

    result = verify_bgp_convergence_from_running_config(tgen, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n" "BGP is converged \n Error : {}".format(
        tc_name, result
    )
    logger.info("Expected behaviour: {}".format(result))

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]

        result = verify_rib(
            tgen, addr_type, "r1", input_dict, protocol="bgp", next_hop=next_hop
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Re-apply redistribution on R3.")

    input_dict_1 = {
        "r3": {
            "bgp": {
                "local_as": "300",
                "address_family": {
                    "ipv4": {
                        "unicast": {"redistribute": [{"redist_type": "connected"}]}
                    },
                    "ipv6": {
                        "unicast": {"redistribute": [{"redist_type": "connected"}]}
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step(
        "Verify that peering between R1-R4 goes down and prefix "
        "of link r3-r4 with NHT R4 is withdrawn."
    )

    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]

        result = verify_rib(
            tgen, addr_type, "r1", input_dict, protocol="bgp", next_hop=next_hop
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Remove redistribution from router R4.")

    input_dict_1 = {
        "r4": {
            "bgp": {
                "local_as": "400",
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "redistribute": [
                                {"redist_type": "connected", "delete": True}
                            ]
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "redistribute": [
                                {"redist_type": "connected", "delete": True}
                            ]
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step(
        "Verify that peering between R1-R4 doesn't go down but prefix "
        "of link r3-r4 with NHT R4 is withdrawn."
    )

    logger.info("Sleeping for holddowntimer: {}".format(HOLD_DOWN_TIMER))
    sleep(HOLD_DOWN_TIMER + 1)

    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r4"]["links"]["r3"][addr_type].split("/")[0]

        result = verify_rib(
            tgen,
            addr_type,
            "r1",
            input_dict,
            protocol="bgp",
            next_hop=next_hop,
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Route is still present \n Error : {}".format(
            tc_name, result
        )

    step("Re-apply redistribution on R4.")

    input_dict_1 = {
        "r4": {
            "bgp": {
                "local_as": "400",
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "redistribute": [
                                {"redist_type": "connected", "delete": True}
                            ]
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "redistribute": [
                                {"redist_type": "connected", "delete": True}
                            ]
                        }
                    },
                },
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify that prefix of link r3-r4 is re-learned via NHT R4.")

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]

        result = verify_rib(
            tgen, addr_type, "r1", input_dict, protocol="bgp", next_hop=next_hop
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Toggle the interface on R3.")

    intf_r3_r4 = topo["routers"]["r3"]["links"]["r4"]["interface"]
    shutdown_bringup_interface(tgen, "r3", intf_r3_r4)

    step(
        "Verify that peering between R1-R4 goes down and comes up when "
        "interface is toggled. Also prefix of link r3-r4(via both NHTs) is"
        " withdrawn and re-learned accordingly."
    )

    result = verify_bgp_convergence_from_running_config(tgen, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n" "BGP is converged \n Error : {}".format(
        tc_name, result
    )
    logger.info("Expected behaviour: {}".format(result))

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r4"]["links"]["r3"][addr_type].split("/")[0]

        result = verify_rib(
            tgen,
            addr_type,
            "r1",
            input_dict,
            protocol="bgp",
            next_hop=next_hop,
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Route is still present \n Error : {}".format(
            tc_name, result
        )

    shutdown_bringup_interface(tgen, "r3", intf_r3_r4, True)

    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]

        result = verify_rib(
            tgen, addr_type, "r1", input_dict, protocol="bgp", next_hop=next_hop
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Toggle the interface on R4.")

    intf_r4_r3 = topo["routers"]["r4"]["links"]["r3"]["interface"]
    shutdown_bringup_interface(tgen, "r4", intf_r4_r3)

    step(
        "Verify that peering between R1-R4 goes down and comes up when"
        "interface is toggled. Also prefix of link r3-r4(via R4)"
        " is withdrawn and re-learned accordingly."
    )

    result = verify_bgp_convergence_from_running_config(tgen, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n" "BGP is converged \n Error : {}".format(
        tc_name, result
    )
    logger.info("Expected behaviour: {}".format(result))

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r4"]["links"]["r3"][addr_type].split("/")[0]

        result = verify_rib(
            tgen,
            addr_type,
            "r1",
            input_dict,
            protocol="bgp",
            next_hop=next_hop,
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Route is still present \n Error : {}".format(
            tc_name, result
        )

    shutdown_bringup_interface(tgen, "r4", intf_r4_r3, True)

    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    for addr_type in ADDR_TYPES:
        routes = {"ipv4": ["10.0.3.0/24"], "ipv6": ["fd00:0:0:3::/64"]}

        input_dict = {"r1": {"static_routes": [{"network": routes[addr_type]}]}}
        next_hop = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]

        result = verify_rib(
            tgen, addr_type, "r1", input_dict, protocol="bgp", next_hop=next_hop
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_BGP_path_attributes_default_values_p1(request):
    """
    Verify that BGP path attributes are present in CLI
    outputs and JSON format, even if set to default.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config: Configure BGP neighborship, between R1-R2 & R1-R3")
    reset_config_on_routers(tgen)

    step("Advertise a set of prefixes from R1 to both peers R2 and R3")
    for addr_type in ADDR_TYPES:
        input_dict_1 = {
            "r1": {
                "static_routes": [{"network": NETWORK[addr_type], "next_hop": "null0"}]
            }
        }
        result = create_static_routes(tgen, input_dict_1)

    input_dict_2 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                    "ipv6": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_2)

    step(
        "Verify that advertised prefixes are received on R4 and well"
        "known attributes are present in the CLI and JSON outputs with"
        "default values without any route-map config."
    )
    for addr_type in ADDR_TYPES:
        input_dict_3 = {"r4": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_bgp_rib(
            tgen,
            addr_type,
            "r4",
            input_dict_3,
            next_hop=[
                topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
                topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
            ],
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r4": {
                "route_maps": {
                    "rmap_pf": [{"set": {"origin": "incomplete", "aspath": "300 100"}}]
                }
            }
        }

        result = verify_bgp_attributes(
            tgen,
            addr_type,
            "r4",
            NETWORK[addr_type],
            rmap_name="rmap_pf",
            input_dict=input_dict_4,
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    step(
        "Configure a route-map to set below attribute value as 500"
        "and apply on R4 in an inbound direction"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r4": {
                "route_maps": {
                    "Path_Attribue": [
                        {
                            "action": "permit",
                            "set": {
                                "path": {"as_num": 500, "as_action": "prepend"},
                                "locPrf": 500,
                                "origin": "egp",
                            },
                        }
                    ]
                }
            }
        }
    result = create_route_maps(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    input_dict_5 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r3": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r3": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step(
        "Verify that once the route-map is applied all the attributes"
        "part of route-map, changes value to 500"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r4": {
                "route_maps": {
                    "rmap_pf": [
                        {
                            "set": {
                                "locPrf": 500,
                                "aspath": "500 300 100",
                                "origin": "EGP",
                            }
                        }
                    ]
                }
            }
        }
        result = verify_bgp_attributes(
            tgen,
            addr_type,
            "r4",
            NETWORK[addr_type],
            rmap_name="rmap_pf",
            input_dict=input_dict_4,
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    step("Remove the route-map from R4")
    input_dict_5 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r3": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                    "delete": True,
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r3": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                    "delete": True,
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step(
        "Verify on R4 that well known attributes are present in the CLI &"
        "JSON outputs again with default values without route-map config"
    )
    for addr_type in ADDR_TYPES:
        input_dict_4 = {
            "r4": {
                "route_maps": {
                    "rmap_pf": [{"set": {"aspath": "300 100", "origin": "incomplete"}}]
                }
            }
        }
        result = verify_bgp_attributes(
            tgen,
            addr_type,
            "r4",
            NETWORK[addr_type],
            rmap_name="rmap_pf",
            input_dict=input_dict_4,
            nexthop=None,
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_BGP_peering_bw_loopback_and_physical_p1(request):
    """
    Verifying the BGP peering between loopback and
    physical link's IP of 2 peer routers.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config :Configure BGP neighborship between R1 and R3.")
    reset_config_on_routers(tgen)

    step("Configure a loopback interface on R1")
    dut = "r1"
    create_interface_in_kernel(
        tgen, dut, "lo10", "1.1.1.1", netmask="255.255.255.255", create=True
    )
    create_interface_in_kernel(
        tgen, dut, "lo10", "1:1::1:1", netmask="128", create=True
    )

    step("Configure BGP session between R1's loopbak & R3")
    for addr_type in ADDR_TYPES:
        input_dict_1 = {
            "r3": {
                "static_routes": [
                    {
                        "network": Loopabck_IP["Lo_R1"][addr_type],
                        "next_hop": topo["routers"]["r1"]["links"]["r3"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )
        result = verify_rib(
            tgen,
            addr_type,
            "r3",
            input_dict_1,
            protocol="static",
            next_hop=topo["routers"]["r1"]["links"]["r3"][addr_type].split("/")[0],
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    for addr_type in ADDR_TYPES:
        raw_config = {
            "r1": {
                "raw_config": [
                    "router bgp {}".format(topo["routers"]["r1"]["bgp"]["local_as"]),
                    "address-family {} unicast".format(addr_type),
                    "neighbor {} update-source lo10".format(
                        topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]
                    ),
                    "neighbor {} timers 1 3".format(
                        topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]
                    ),
                ]
            },
            "r3": {
                "raw_config": [
                    "router bgp {}".format(topo["routers"]["r3"]["bgp"]["local_as"]),
                    "address-family {} unicast".format(addr_type),
                    "no neighbor {} remote-as {}".format(
                        topo["routers"]["r1"]["links"]["r3"][addr_type].split("/")[0],
                        topo["routers"]["r1"]["bgp"]["local_as"],
                    ),
                    "neighbor {} remote-as {}".format(
                        Loopabck_IP["Lo_R1"][addr_type].split("/")[0],
                        topo["routers"]["r1"]["bgp"]["local_as"],
                    ),
                    "neighbor {} ebgp-multihop 3".format(
                        Loopabck_IP["Lo_R1"][addr_type].split("/")[0]
                    ),
                ]
            },
        }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    for addr_type in ADDR_TYPES:
        if addr_type == "ipv6":
            raw_config = {
                "r3": {
                    "raw_config": [
                        "router bgp {}".format(
                            topo["routers"]["r3"]["bgp"]["local_as"]
                        ),
                        "address-family {} unicast".format(addr_type),
                        "neighbor {} activate".format(
                            Loopabck_IP["Lo_R1"][addr_type].split("/")[0]
                        ),
                    ]
                }
            }
        else:
            raw_config = {
                "r3": {
                    "raw_config": [
                        "router bgp {}".format(
                            topo["routers"]["r3"]["bgp"]["local_as"]
                        ),
                        "address-family {} unicast".format(addr_type),
                        "no neighbor {} activate".format(
                            Loopabck_IP["Lo_R1"]["ipv6"].split("/")[0]
                        ),
                    ]
                }
            }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    step("Verify that BGP neighborship between R1 and R3 comes up")
    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Remove ebgp-multihop command from R3")
    for addr_type in ADDR_TYPES:
        raw_config = {
            "r3": {
                "raw_config": [
                    "router bgp {}".format(topo["routers"]["r3"]["bgp"]["local_as"]),
                    "no neighbor {} ebgp-multihop 3".format(
                        Loopabck_IP["Lo_R1"][addr_type].split("/")[0]
                    ),
                ]
            }
        }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    step("Verify that once eBGP multi-hop is removed, BGP session goes down")
    result = verify_bgp_convergence_from_running_config(tgen, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "BGP is converged \n Error: {}".format(
        tc_name, result
    )

    step("Add ebgp-multihop command on R3 again")
    for addr_type in ADDR_TYPES:
        raw_config = {
            "r3": {
                "raw_config": [
                    "router bgp {}".format(topo["routers"]["r3"]["bgp"]["local_as"]),
                    "neighbor {} ebgp-multihop 3".format(
                        Loopabck_IP["Lo_R1"][addr_type].split("/")[0]
                    ),
                ]
            }
        }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    step("Verify that BGP neighborship between R1 and R3 comes up")
    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Remove update-source command from R1")
    for addr_type in ADDR_TYPES:
        raw_config = {
            "r1": {
                "raw_config": [
                    "router bgp {}".format(topo["routers"]["r1"]["bgp"]["local_as"]),
                    "no neighbor {} update-source lo10".format(
                        topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]
                    ),
                ]
            }
        }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    step("Verify that BGP session goes down, when update-source is removed")
    result = verify_bgp_convergence_from_running_config(tgen, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "BGP is converged \n Error: {}".format(
        tc_name, result
    )

    step("Add update-source command on R1 again")
    for addr_type in ADDR_TYPES:
        raw_config = {
            "r1": {
                "raw_config": [
                    "router bgp {}".format(topo["routers"]["r1"]["bgp"]["local_as"]),
                    "neighbor {} update-source lo10".format(
                        topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]
                    ),
                ]
            }
        }
        result = apply_raw_config(tgen, raw_config)
        assert result is True, "Testcase {} : Failed Error : {}".format(tc_name, result)

    step("Verify that BGP neighborship between R1 and R3 comes up")
    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Remove static route from R3")
    for addr_type in ADDR_TYPES:
        input_dict_1 = {
            "r3": {
                "static_routes": [
                    {
                        "network": Loopabck_IP["Lo_R1"][addr_type],
                        "next_hop": topo["routers"]["r1"]["links"]["r3"][
                            addr_type
                        ].split("/")[0],
                        "delete": True,
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )
        result = verify_rib(
            tgen,
            addr_type,
            "r3",
            input_dict_1,
            protocol="static",
            next_hop=topo["routers"]["r1"]["links"]["r3"][addr_type].split("/")[0],
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Routes are still present \n Error: {}".format(
            tc_name, result
        )

    sleep(3)
    step("Verify that BGP session goes down, when static route is removed")
    result = verify_bgp_convergence_from_running_config(tgen, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "BGP is converged \n Error: {}".format(
        tc_name, result
    )

    step("Add static route on R3 again")
    for addr_type in ADDR_TYPES:
        input_dict_1 = {
            "r3": {
                "static_routes": [
                    {
                        "network": Loopabck_IP["Lo_R1"][addr_type],
                        "next_hop": topo["routers"]["r1"]["links"]["r3"][
                            addr_type
                        ].split("/")[0],
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )
        result = verify_rib(
            tgen,
            addr_type,
            "r3",
            input_dict_1,
            protocol="static",
            next_hop=topo["routers"]["r1"]["links"]["r3"][addr_type].split("/")[0],
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Verify that BGP neighborship between R1 and R3 comes up")
    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Toggle physical interface on R1")
    intf_r1_r3 = topo["routers"]["r1"]["links"]["r3"]["interface"]
    shutdown_bringup_interface(tgen, "r1", intf_r1_r3)
    sleep(3)
    step("Verify that BGP neighborship between R1 and R3 goes down")
    result = verify_bgp_convergence_from_running_config(tgen, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "BGP is converged \n Error: {}".format(
        tc_name, result
    )

    intf_r1_r3 = topo["routers"]["r1"]["links"]["r3"]["interface"]
    shutdown_bringup_interface(tgen, "r1", intf_r1_r3, True)

    step("Verify that BGP neighborship between R1 and R3 comes up")
    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_BGP_active_standby_preemption_and_ecmp_p1(request):
    """
    Verify that BGP Active/Standby/Pre-emption/ECMP.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config :Configure BGP neighborship between R1 and R3.")
    reset_config_on_routers(tgen)

    step("Change the AS number on R2 as 200")
    input_dict = {"r2": {"bgp": {"local_as": 200}}}
    result = modify_as_number(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify BGP converge after changing the AS number on R2")
    result = verify_bgp_convergence_from_running_config(tgen)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Advertise a set of prefixes from R1 to both peers R2 & R3")
    for addr_type in ADDR_TYPES:
        input_dict_1 = {
            "r1": {
                "static_routes": [{"network": NETWORK[addr_type], "next_hop": "null0"}]
            }
        }
        result = create_static_routes(tgen, input_dict_1)
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    input_dict_2 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                    "ipv6": {"unicast": {"redistribute": [{"redist_type": "static"}]}},
                }
            }
        }
    }

    result = create_router_bgp(tgen, topo, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify that R4 receives BGP prefixes via both peer routers R2 & R3")
    for addr_type in ADDR_TYPES:
        input_dict_3 = {"r4": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_bgp_rib(
            tgen,
            addr_type,
            "r4",
            input_dict_3,
            next_hop=[
                topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
                topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
            ],
        )
        assert result is True, "Testcase {}: Failed \n Error : {}".format(
            tc_name, result
        )

    step(
        "Configure a route-map to set as-path attribute and"
        "apply on R3 in an inbound direction:"
    )

    input_dict_4 = {
        "r3": {
            "route_maps": {
                "Path_Attribue": [
                    {
                        "action": "permit",
                        "set": {"path": {"as_num": 123, "as_action": "prepend"}},
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    input_dict_5 = {
        "r3": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify on R4, BGP routes with shorter as-path are installed in FIB")
    for addr_type in ADDR_TYPES:
        dut = "r4"
        protocol = "bgp"
        input_dict_6 = {"r4": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_rib(
            tgen,
            addr_type,
            dut,
            input_dict_6,
            next_hop=topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
            protocol=protocol,
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Shutdown BGP neighorship between R1-R2")
    dut = "r4"
    intf_r4_r2 = topo["routers"]["r4"]["links"]["r2"]["interface"]
    shutdown_bringup_interface(tgen, dut, intf_r4_r2)

    step(
        "Verify that prefixes from next-hop via R2 are withdrawn"
        "and installed via next-hop as R3"
    )
    result = verify_rib(
        tgen,
        addr_type,
        dut,
        input_dict_2,
        next_hop=topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
        protocol=protocol,
    )
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Do a no shut for BGP neighorship between R2-R4")
    shutdown_bringup_interface(tgen, dut, intf_r4_r2, ifaceaction=True)

    step(
        "Verify that prefixes from next-hop via R3 are withdrawn"
        "from R4 and installed via next-hop as R2 (preemption)"
    )
    result = verify_rib(
        tgen,
        addr_type,
        dut,
        input_dict_2,
        next_hop=topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
        protocol=protocol,
    )
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Remove the route-map from R3's neighbor statement")
    input_dict_5 = {
        "r3": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                    "delete": True,
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r1": {
                                    "dest_link": {
                                        "r3": {
                                            "route_maps": [
                                                {
                                                    "name": "Path_Attribue",
                                                    "direction": "in",
                                                    "delete": True,
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Configure multipath-relax and maximum-paths 2 on R4 for ECMP")
    input_dict_8 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {"unicast": {"maximum_paths": {"ebgp": 2}}},
                    "ipv6": {"unicast": {"maximum_paths": {"ebgp": 2}}},
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_8)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    maxpath_relax = {
        "r4": {"bgp": {"local_as": "400", "bestpath": {"aspath": "multipath-relax"}}}
    }

    result = create_router_bgp(tgen, topo, maxpath_relax)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify FIB of R4, BGP prefixes with ECMP next-hop via R2 and R3")
    for addr_type in ADDR_TYPES:
        input_dict = {"r4": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_rib(
            tgen,
            addr_type,
            "r4",
            input_dict,
            next_hop=[
                topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
                topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
            ],
        )
        assert result is True, "Testcase {} : Failed \n Error : {}".format(
            tc_name, result
        )

    step("Remove multipath-relax command from R4")

    del_maxpath_relax = {
        "r4": {
            "bgp": {
                "local_as": "400",
                "bestpath": {"aspath": "multipath-relax", "delete": True},
            }
        }
    }

    result = create_router_bgp(tgen, topo, del_maxpath_relax)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify that ECMP is no longer happening on R4.")
    for addr_type in ADDR_TYPES:
        input_dict = {"r4": {"static_routes": [{"network": NETWORK[addr_type]}]}}
        result = verify_rib(
            tgen,
            addr_type,
            "r4",
            input_dict,
            next_hop=[
                topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
                topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
            ],
            expected=False,
        )
        assert (
            result is not True
        ), "Testcase {} : Failed \n " "Routes are still present \n Error: {}".format(
            tc_name, result
        )

    step("Reconfigure multipath-relax command on R4")
    result = create_router_bgp(tgen, topo, maxpath_relax)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify FIB of R4, BGP prefixes with ECMP next-hop via R2 and R3")
    result = verify_rib(
        tgen,
        addr_type,
        "r4",
        input_dict,
        next_hop=[
            topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
            topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
        ],
    )
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Remove maximum-path 2 command from R4")
    input_dict_8 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "maximum_paths": {
                                "ebgp": 1,
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "maximum_paths": {
                                "ebgp": 1,
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_8)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify that ECMP is no longer happening on R4")
    result = verify_rib(
        tgen,
        addr_type,
        "r4",
        input_dict,
        next_hop=[
            topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
            topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
        ],
        expected=False,
    )
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "Routes are still present \n Error: {}".format(
        tc_name, result
    )

    step("Re-configure maximum-path 2 command on R4")
    input_dict_8 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "maximum_paths": {
                                "ebgp": 2,
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "maximum_paths": {
                                "ebgp": 2,
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_8)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Verify FIB of R4, BGP prefixes with ECMP next-hop via R2 and R3")
    result = verify_rib(
        tgen,
        addr_type,
        "r4",
        input_dict,
        next_hop=[
            topo["routers"]["r2"]["links"]["r4"][addr_type].split("/")[0],
            topo["routers"]["r3"]["links"]["r4"][addr_type].split("/")[0],
        ],
    )
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_password_authentication_for_eBGP_and_iBGP_peers_p1(request):
    """
    Verify password authentication for eBGP and iBGP peers.
    """

    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Initial config :Configure BGP neighborship between R1 and R3.")
    reset_config_on_routers(tgen)

    step(
        "Add a static route on R1 for loopbacks IP's reachability of R2, R3"
        "and on R2 and R3 for loopback IP of R1"
    )
    for addr_type in ADDR_TYPES:
        nh1 = topo["routers"]["r3"]["links"]["r1"][addr_type].split("/")[0]
        nh2 = topo["routers"]["r1"]["links"]["r2"][addr_type].split("/")[0]
        nh3 = topo["routers"]["r1"]["links"]["r3"][addr_type].split("/")[0]
        nh4 = topo["routers"]["r2"]["links"]["r1"][addr_type].split("/")[0]
        input_dict_1 = {
            "r1": {
                "static_routes": [
                    {
                        "network": topo["routers"]["r3"]["links"]["lo"][addr_type],
                        "next_hop": nh1,
                    }
                ]
            }
        }
        input_dict_2 = {
            "r2": {
                "static_routes": [
                    {
                        "network": topo["routers"]["r1"]["links"]["lo"][addr_type],
                        "next_hop": nh2,
                    }
                ]
            }
        }
        input_dict_3 = {
            "r3": {
                "static_routes": [
                    {
                        "network": topo["routers"]["r1"]["links"]["lo"][addr_type],
                        "next_hop": nh3,
                    }
                ]
            }
        }
        input_dict_4 = {
            "r1": {
                "static_routes": [
                    {
                        "network": topo["routers"]["r2"]["links"]["lo"][addr_type],
                        "next_hop": nh4,
                    }
                ]
            }
        }
        dut_list = ["r1", "r2", "r3", "r1"]
        nexthop_list = [nh1, nh2, nh3, nh4]
        input_dict_list = [input_dict_1, input_dict_2, input_dict_3, input_dict_4]
        for dut, next_hop, input_dict in zip(dut_list, nexthop_list, input_dict_list):
            result = create_static_routes(tgen, input_dict)
            assert result is True, "Testcase {} : Failed \n Error : {}".format(
                tc_name, result
            )

            step("Verify that static routes are installed in FIB of routers")
            result = verify_rib(
                tgen, addr_type, dut, input_dict, next_hop=next_hop, protocol="static"
            )
            assert result is True, "Testcase {} : Failed \n Error : {}".format(
                tc_name, result
            )

    step("Configure BGP sessions between R1-R2 and R1-R3 over loopback IPs")
    for routerN in ["r1", "r3"]:
        for addr_type in ADDR_TYPES:
            if routerN == "r1":
                bgp_neighbor = "r3"
            elif routerN == "r3":
                bgp_neighbor = "r1"
            topo["routers"][routerN]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ][bgp_neighbor]["dest_link"] = {
                "lo": {"ebgp_multihop": 2, "source_link": "lo"}
            }
    build_config_from_json(tgen, topo, save_bkup=False)

    for routerN in ["r1", "r2"]:
        for addr_type in ADDR_TYPES:
            if routerN == "r1":
                bgp_neighbor = "r2"
            elif routerN == "r2":
                bgp_neighbor = "r1"
            topo["routers"][routerN]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ][bgp_neighbor]["dest_link"] = {"lo": {"source_link": "lo"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    for routerN in ["r1", "r2", "r3"]:
        for addr_type in ADDR_TYPES:
            for bgp_neighbor in topo["routers"][routerN]["bgp"]["address_family"][
                addr_type
            ]["unicast"]["neighbor"].keys():
                if routerN in ["r1", "r2", "r3"] and bgp_neighbor == "r4":
                    continue
                if addr_type == "ipv4":
                    topo["routers"][routerN]["bgp"]["address_family"][addr_type][
                        "unicast"
                    ]["neighbor"][bgp_neighbor]["dest_link"] = {
                        "lo": {"deactivate": "ipv6"}
                    }
                elif addr_type == "ipv6":
                    topo["routers"][routerN]["bgp"]["address_family"][addr_type][
                        "unicast"
                    ]["neighbor"][bgp_neighbor]["dest_link"] = {
                        "lo": {"deactivate": "ipv4"}
                    }
    build_config_from_json(tgen, topo, save_bkup=False)

    result = verify_bgp_convergence(tgen, topo)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Configure authentication password on R1 for neighbor statements")
    for bgp_neighbor in ["r2", "r3"]:
        for addr_type in ADDR_TYPES:
            topo["routers"]["r1"]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ][bgp_neighbor]["dest_link"] = {"lo": {"password": "vmware"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    step(
        "Verify that both sessions go down as only R1 has password"
        "configured but not peer routers"
    )
    result = verify_bgp_convergence(tgen, topo, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "BGP is converged \n Error: {}".format(
        tc_name, result
    )

    step("configure same password on R2 and R3")
    for routerN in ["r2", "r3"]:
        for addr_type in ADDR_TYPES:
            topo["routers"][routerN]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ]["r1"]["dest_link"] = {"lo": {"password": "vmware"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    step("Verify that all BGP sessions come up due to identical passwords")
    result = verify_bgp_convergence(tgen, topo)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Configure same password on R2 and R3, but in CAPs.")
    for routerN in ["r2", "r3"]:
        for addr_type in ADDR_TYPES:
            topo["routers"][routerN]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ]["r1"]["dest_link"] = {"lo": {"password": "VMWARE"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    step(
        "Verify that BGP sessions do not come up as password"
        "strings are in CAPs on R2 and R3"
    )
    result = verify_bgp_convergence(tgen, topo, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "BGP is converged \n Error: {}".format(
        tc_name, result
    )

    step("Configure same password on R2 and R3 without CAPs")
    for routerN in ["r2", "r3"]:
        for addr_type in ADDR_TYPES:
            topo["routers"][routerN]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ]["r1"]["dest_link"] = {"lo": {"password": "vmware"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    step("Verify all BGP sessions come up again due to identical passwords")
    result = verify_bgp_convergence(tgen, topo)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    step("Remove password from R1")
    for bgp_neighbor in ["r2", "r3"]:
        for addr_type in ADDR_TYPES:
            topo["routers"]["r1"]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ][bgp_neighbor]["dest_link"] = {"lo": {"no_password": "vmware"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    step("Verify if password is removed from R1, both sessions go down again")
    result = verify_bgp_convergence(tgen, topo, expected=False)
    assert (
        result is not True
    ), "Testcase {} : Failed \n " "BGP is converged \n Error: {}".format(
        tc_name, result
    )

    step("Configure alphanumeric password on R1 and peer routers R2,R3")
    for bgp_neighbor in ["r2", "r3"]:
        for addr_type in ADDR_TYPES:
            topo["routers"]["r1"]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ][bgp_neighbor]["dest_link"] = {"lo": {"password": "Vmware@123"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    for routerN in ["r2", "r3"]:
        for addr_type in ADDR_TYPES:
            topo["routers"][routerN]["bgp"]["address_family"][addr_type]["unicast"][
                "neighbor"
            ]["r1"]["dest_link"] = {"lo": {"password": "Vmware@123"}}
    build_config_from_json(tgen, topo, save_bkup=False)

    step(
        "Verify that sessions Come up irrespective of characters"
        "used in password string"
    )
    result = verify_bgp_convergence(tgen, topo)
    assert result is True, "Testcase {} : Failed \n Error : {}".format(tc_name, result)

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: ISC
#
# March 21 2024, Igor Ryzhov <iryzhov@nfware.com>
#
# Copyright (c) 2024, NFWare Inc.
#

"""
Test YANG Notifications
"""
import json
import os
import threading

import pytest
from lib.common_config import retry
from lib.topogen import Topogen
from lib.topotest import json_cmp

pytestmark = [pytest.mark.ripd, pytest.mark.mgmtd]

CWD = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture(scope="module")
def tgen(request):
    "Setup/Teardown the environment and provide tgen argument to tests"

    topodef = {"s1": ("r1",)}

    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for _, router in router_list.items():
        router.load_frr_config("frr.conf")

    tgen.start_router()
    yield tgen
    tgen.stop_topology()


# Verify the backend test client has connected
@retry(retry_timeout=10)
def check_client_connect(r1):
    out = r1.vtysh_cmd("show mgmt backend-adapter all")
    assert "mgmtd-testc" in out


def test_backend_rpc(tgen):
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]

    # Run the backend test client which registers to handle the `clear ip rip` command.
    be_client_path = "/usr/lib/frr/mgmtd_testc"
    rc, _, _ = r1.net.cmd_status(be_client_path + " --help")

    if rc:
        pytest.skip("No mgmtd_testc")

    out = []

    def run_testc():
        output = r1.net.cmd_raises(
            be_client_path + " --timeout 10 --log file:mgmt_testc.log"
        )
        out.append(json.loads(output))

    t = threading.Thread(target=run_testc)
    t.start()

    # We need to wait for mgmtd_testc to connect before issuing the command.
    res = check_client_connect(r1)
    assert res is None

    r1.vtysh_cmd("clear ip rip vrf testname")

    t.join()

    jsout = out[0]

    expected = {"frr-ripd:clear-rip-route": {"vrf": "testname"}}
    result = json_cmp(jsout, expected)
    assert result is None



#!/usr/bin/env python
# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: ISC
#
# Copyright (c) 2021, LabN Consulting, L.L.C.
# Copyright (c) 2019-2020 by
# Donatas Abraitis <donatas.abraitis@gmail.com>
#
# noqa: E501
#
"""
Test static route functionality
"""
import pytest
from lib.common_config import step
from lib.topogen import Topogen
from munet.watchlog import WatchLog

pytestmark = [pytest.mark.staticd, pytest.mark.mgmtd]


@pytest.fixture(scope="module")
def tgen(request):
    "Setup/Teardown the environment and provide tgen argument to tests"

    topodef = {"s1": ("r1",)}
    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()

    for _, router in tgen.routers().items():
        router.load_frr_config("frr.conf")

    tgen.start_router()
    yield tgen
    tgen.stop_topology()


def test_client_debug_enable(tgen):
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]

    # Start watching log files
    watch_mgmtd_log = WatchLog(r1.net.rundir / "mgmtd.log")
    watch_staticd_log = WatchLog(r1.net.rundir / "staticd.log")

    def __test_debug(r1, on):
        watch_mgmtd_log.snapshot()
        watch_staticd_log.snapshot()

        # Add ip route and remove look for debug.
        r1.vtysh_cmd("conf t\nip route 11.11.11.11/32 1.1.1.2")
        r1.vtysh_cmd("conf t\nno ip route 11.11.11.11/32 1.1.1.2")

        new_mgmt_logs = watch_mgmtd_log.snapshot()
        new_be_logs = watch_staticd_log.snapshot()

        fe_cl_msg = "Sending SET_CONFIG_REQ"
        fe_ad_msg = "Got SETCFG_REQ"
        be_ad_msg = "Sending CFGDATA_CREATE_REQ"
        be_cl_msg = "Got CFG_APPLY_REQ"

        if on:
            assert fe_cl_msg in new_mgmt_logs
            assert fe_ad_msg in new_mgmt_logs
            assert be_ad_msg in new_mgmt_logs
            assert be_cl_msg in new_be_logs
        else:
            assert fe_cl_msg not in new_mgmt_logs
            assert fe_ad_msg not in new_mgmt_logs
            assert be_ad_msg not in new_mgmt_logs
            assert be_cl_msg not in new_be_logs

    step("test debug off")
    __test_debug(r1, False)

    # Turn it on
    step("tests debug on")
    r1.vtysh_cmd("debug mgmt client frontend")
    r1.vtysh_cmd("debug mgmt client backend")
    r1.vtysh_cmd("debug mgmt backend frontend")
    __test_debug(r1, True)

    # Turn it off
    step("tests debug off")
    r1.vtysh_cmd("no debug mgmt client frontend")
    r1.vtysh_cmd("no debug mgmt client backend")
    r1.vtysh_cmd("no debug mgmt backend frontend")
    __test_debug(r1, False)

    # Configure it on
    step("tests debug on")
    r1.vtysh_cmd("conf t\ndebug mgmt client frontend")
    r1.vtysh_cmd("conf t\ndebug mgmt client backend")
    r1.vtysh_cmd("conf t\ndebug mgmt backend frontend")
    __test_debug(r1, True)

    # Configure it off
    step("tests debug off")
    r1.vtysh_cmd("conf t\nno debug mgmt client frontend")
    r1.vtysh_cmd("conf t\nno debug mgmt client backend")
    r1.vtysh_cmd("conf t\nno debug mgmt backend frontend")
    __test_debug(r1, False)

    # Turn it on
    step("tests debug on")
    r1.vtysh_cmd("debug mgmt client frontend")
    r1.vtysh_cmd("debug mgmt client backend")
    r1.vtysh_cmd("debug mgmt backend frontend")
    __test_debug(r1, True)
    # Configure it on
    step("tests debug on")
    r1.vtysh_cmd("conf t\ndebug mgmt client frontend")
    r1.vtysh_cmd("conf t\ndebug mgmt client backend")
    r1.vtysh_cmd("conf t\ndebug mgmt backend frontend")
    __test_debug(r1, True)
    # Turn it off
    step("tests debug on")
    r1.vtysh_cmd("no debug mgmt client frontend")
    r1.vtysh_cmd("no debug mgmt client backend")
    r1.vtysh_cmd("no debug mgmt backend frontend")
    __test_debug(r1, True)
    # Configure it off
    step("tests debug off")
    r1.vtysh_cmd("conf t\nno debug mgmt client frontend")
    r1.vtysh_cmd("conf t\nno debug mgmt client backend")
    r1.vtysh_cmd("conf t\nno debug mgmt backend frontend")
    __test_debug(r1, False)



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2023 by Louis Scalbert <louis.scalbert@6wind.com>
# Copyright 2023 6WIND S.A.
#

"""

"""

import os
import re
import sys
import json
import pytest
import functools

from copy import deepcopy

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.common_config import kill_router_daemons, start_router_daemons, step

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    for rtr in [1, 2]:
        tgen.add_router("r{}".format(rtr))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for rtr in [1, 2]:
        tgen.gears["r{}".format(rtr)].cmd("ip link add vrf1 type vrf table 10")
        tgen.gears["r{}".format(rtr)].cmd("ip link set vrf1 up")
        tgen.gears["r{}".format(rtr)].cmd(
            "ip address add dev vrf1 192.0.3.{}/32".format(rtr)
        )
        tgen.gears["r{}".format(rtr)].run(
            "sysctl -w net.mpls.conf.r{}-eth0.input=1".format(rtr)
        )
        tgen.gears["r{}".format(rtr)].run("sysctl -w net.mpls.conf.vrf1.input=1")

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_STATIC, os.path.join(CWD, "{}/staticd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_LDP, os.path.join(CWD, "{}/ldpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def check_bgp_vpn_prefix(label, rname="r1", rd=None):
    tgen = get_topogen()

    if rd:
        output = json.loads(
            tgen.gears[rname].vtysh_cmd(
                "show bgp ipv4 vpn rd {} 192.0.3.2/32 json".format(rd)
            )
        )
    else:
        output = json.loads(
            tgen.gears[rname].vtysh_cmd(
                "show bgp vrf vrf1 ipv4 unicast 192.0.3.2/32 json"
            )
        )

    if label == "auto":
        expected = {
            "paths": [
                {
                    "valid": True,
                    "aspath": {"string": "65002"},
                    "nexthops": [{"ip": "192.0.2.2"}],
                },
            ]
        }
    elif label and not rd:
        expected = {
            "paths": [
                {
                    "valid": True,
                    "remoteLabel": label,
                    "aspath": {"string": "65002"},
                    "nexthops": [{"ip": "192.0.2.2"}],
                },
            ]
        }
    elif label and rd:
        expected = {
            "102:1": {
                "prefix": "192.0.3.2/32",
                "paths": [
                    {
                        "valid": True,
                        "remoteLabel": label,
                        "nexthops": [{"ip": "0.0.0.0"}],
                    }
                ],
            }
        }
    else:
        expected = {}

    return topotest.json_cmp(output, expected, exact=(label is None))


def check_mpls_table(label, protocol):
    tgen = get_topogen()

    if label == "auto":
        cmd = "show mpls table json"
    else:
        cmd = "show mpls table {} json".format(label)

    output = json.loads(tgen.gears["r2"].vtysh_cmd(cmd))

    if label == "auto" and protocol:
        output_copy = deepcopy(output)
        for _, data in output_copy.items():
            for nexthop in data.get("nexthops", []):
                if nexthop.get("type", None) != protocol:
                    continue
                output = data
                break

    if protocol:
        expected = {
            "nexthops": [
                {
                    "type": protocol,
                },
            ]
        }
    else:
        expected = {}

    return topotest.json_cmp(output, expected, exact=(protocol is None))


def check_mpls_ldp_binding():
    tgen = get_topogen()

    output = json.loads(
        tgen.gears["r1"].vtysh_cmd("show mpls ldp binding 192.0.2.2/32 json")
    )
    expected = {
        "bindings": [
            {
                "prefix": "192.0.2.2/32",
                "localLabel": "16",  # first available label
                "inUse": 1,
            },
        ]
    }

    return topotest.json_cmp(output, expected)


def test_convergence():
    "Test protocol convergence"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    step("Check BGP and LDP convergence")
    test_func = functools.partial(check_bgp_vpn_prefix, 2222)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP prefix on R1"

    test_func = functools.partial(check_mpls_ldp_binding)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP binding on R2"

    test_func = functools.partial(check_mpls_table, 16, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    test_func = functools.partial(check_mpls_table, 2222, "BGP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP label on R2"

    output = tgen.net["r2"].cmd("vtysh -c 'show debugging label-table' | grep Proto")
    assert re.match(
        r"Proto ldp: \[16/(1[7-9]|[2-9]\d+|\d{3,})\]", output
    ), "Failed to see LDP label chunk"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: [2222/2222]" in output, "Failed to see BGP label chunk"


def test_vpn_label_export_16():
    "Test that assigning the label value of 16 is not possible because it used by LDP"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "label vpn export 16"
    )

    step("Check that label vpn export 16 fails")
    test_func = functools.partial(check_bgp_vpn_prefix, None)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Unexpected BGP prefix on R1"

    test_func = functools.partial(check_mpls_ldp_binding)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP binding on R2"

    test_func = functools.partial(check_mpls_table, 16, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    test_func = functools.partial(check_mpls_table, 2222, None)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Unexpected BGP label on R2"

    output = tgen.net["r2"].cmd("vtysh -c 'show debugging label-table' | grep Proto")
    assert re.match(
        r"Proto ldp: \[16/(1[7-9]|[2-9]\d+|\d{3,})\]", output
    ), "Failed to see LDP label chunk"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp" not in output, "Unexpected BGP label chunk"


def test_vpn_label_export_2222():
    "Test that setting back the label value of 2222 works"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "label vpn export 2222"
    )

    step("Check that label vpn export 2222 is OK")
    test_func = functools.partial(check_bgp_vpn_prefix, 2222)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP prefix on R1"

    test_func = functools.partial(check_mpls_ldp_binding)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP binding on R2"

    test_func = functools.partial(check_mpls_table, 16, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    test_func = functools.partial(check_mpls_table, "auto", "BGP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Unexpected BGP label on R2"

    output = tgen.net["r2"].cmd("vtysh -c 'show debugging label-table' | grep Proto")
    assert re.match(
        r"Proto ldp: \[16/(1[7-9]|[2-9]\d+|\d{3,})\]", output
    ), "Failed to see LDP label chunk"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: [2222/2222]" in output, "Failed to see BGP label chunk"


def test_vpn_label_export_auto():
    "Test that setting label vpn export auto works"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "label vpn export auto"
    )

    step("Check that label vpn export auto is OK")
    test_func = functools.partial(check_bgp_vpn_prefix, "auto")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP prefix on R1"

    test_func = functools.partial(check_mpls_ldp_binding)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP binding on R2"

    test_func = functools.partial(check_mpls_table, 16, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    test_func = functools.partial(check_mpls_table, "auto", "BGP")
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert result is None, "Failed to see BGP label on R2"

    output = tgen.net["r2"].cmd("vtysh -c 'show debugging label-table' | grep Proto")
    assert re.match(
        r"Proto ldp: \[16/(1[7-9]|[2-9]\d+|\d{3,})\]", output
    ), "Failed to see LDP label chunk"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: " in output, "Failed to see BGP label chunk"


def test_vpn_label_export_no_auto():
    "Test that UNsetting label vpn export auto removes the prefix from R1 table and R2 LDP table"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    output = json.loads(
        tgen.gears["r1"].vtysh_cmd("show bgp vrf vrf1 ipv4 unicast 192.0.3.2/32 json")
    )

    auto_label = output.get("paths")[0].get("remoteLabel", None)
    assert auto_label is not None, "Failed to fetch prefix label on R1"

    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "no label vpn export auto"
    )

    step("Check that no label vpn export auto is OK")
    test_func = functools.partial(check_bgp_vpn_prefix, 3, rname="r2", rd="102:1")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Unexpected BGP prefix on R2"

    test_func = functools.partial(check_mpls_ldp_binding)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP binding on R2"

    test_func = functools.partial(check_mpls_table, 16, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    test_func = functools.partial(check_mpls_table, auto_label, None)
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert result is None, "Unexpected BGP label on R2"

    output = tgen.net["r2"].cmd("vtysh -c 'show debugging label-table' | grep Proto")
    assert re.match(
        r"Proto ldp: \[16/(1[7-9]|[2-9]\d+|\d{3,})\]", output
    ), "Failed to see LDP label chunk"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: " not in output, "Unexpected BGP label chunk"


def test_vpn_label_export_auto_back():
    "Test that setting back label vpn export auto works"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    output = json.loads(
        tgen.gears["r2"].vtysh_cmd("show bgp vrf vrf1 ipv4 unicast 192.0.3.2/32 json")
    )

    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "label vpn export auto"
    )

    step("Check that label vpn export auto is OK")
    test_func = functools.partial(check_bgp_vpn_prefix, "auto")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP prefix on R1"

    test_func = functools.partial(check_mpls_ldp_binding)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP binding on R2"

    test_func = functools.partial(check_mpls_table, 16, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    test_func = functools.partial(check_mpls_table, "auto", "BGP")
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert result is None, "Failed to see BGP label on R2"

    output = tgen.net["r2"].cmd("vtysh -c 'show debugging label-table' | grep Proto")
    assert re.match(
        r"Proto ldp: \[16/(1[7-9]|[2-9]\d+|\d{3,})\]", output
    ), "Failed to see LDP label chunk"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: " in output, "Failed to see BGP label chunk"


def test_vpn_label_export_manual_from_auto():
    "Test that setting a manual label value from the BGP chunk range works"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    output = json.loads(
        tgen.gears["r1"].vtysh_cmd("show bgp vrf vrf1 ipv4 unicast 192.0.3.2/32 json")
    )

    auto_label = output.get("paths")[0].get("remoteLabel", None)
    assert auto_label is not None, "Failed to fetch prefix label on R1"

    auto_label = auto_label + 1

    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "label vpn export {}".format(auto_label)
    )

    step("Check that label vpn export {} is OK".format(auto_label))
    test_func = functools.partial(
        check_bgp_vpn_prefix, auto_label, rname="r2", rd="102:1"
    )
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see BGP prefix on R2"

    test_func = functools.partial(check_mpls_ldp_binding)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP binding on R2"

    test_func = functools.partial(check_mpls_table, 16, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    test_func = functools.partial(check_mpls_table, auto_label, "BGP")
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert result is None, "Failed to see BGP label on R2"

    output = tgen.net["r2"].cmd("vtysh -c 'show debugging label-table' | grep Proto")
    assert re.match(
        r"Proto ldp: \[16/(1[7-9]|[2-9]\d+|\d{3,})\]", output
    ), "Failed to see LDP label chunk"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: " in output, "Failed to see BGP label chunk"


def test_vpn_label_configure_dynamic_range():
    "Test that if a dynamic range is configured, then the next dynamic allocations will be done in that block"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    tgen.gears["r2"].vtysh_cmd("conf\n" "mpls label dynamic-block 500 1000\n")
    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "label vpn export auto"
    )
    step("Check that label vpn export auto starting at 500 is OK")
    test_func = functools.partial(check_bgp_vpn_prefix, 500, rname="r2", rd="102:1")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Unexpected BGP prefix on R2"

    test_func = functools.partial(check_mpls_table, 500, "BGP")
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert result is None, "Unexpected BGP label on R2"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: " in output, "Failed to see BGP label chunk"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))


def test_vpn_label_restart_ldp():
    "Test that if a dynamic range is configured, then when LDP restarts, it follows the new dynamic range"

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router_list = tgen.routers()

    step("Kill LDP on R2")
    kill_router_daemons(tgen, "r2", ["ldpd"])

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto ldp: " not in output, "Unexpected LDP label chunk"

    step("Bring up LDP on R2")

    start_router_daemons(tgen, "r2", ["ldpd"])

    test_func = functools.partial(check_mpls_table, 628, "LDP")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Failed to see LDP label on R2"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto ldp: [628/691]" in output, "Failed to see LDP label chunk [628/691]"
    assert "Proto ldp: [692/755]" in output, "Failed to see LDP label chunk [692/755]"


def test_vpn_label_unconfigure_dynamic_range():
    "Test that if the dynamic range is unconfigured, then the next dynamic allocations will be done at the first free place."

    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    tgen.gears["r2"].vtysh_cmd("conf\n" "no mpls label dynamic-block 500 1000\n")
    step("Check that unconfiguring label vpn export auto will remove BGP label chunk")
    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "no label vpn export auto"
    )

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: " not in output, "Unexpected BGP label chunk"

    tgen.gears["r2"].vtysh_cmd(
        "conf\n"
        "router bgp 65002 vrf vrf1\n"
        "address-family ipv4 unicast\n"
        "label vpn export auto"
    )
    step("Check that label vpn export auto starting at 16 is OK")
    test_func = functools.partial(check_bgp_vpn_prefix, 16, rname="r2", rd="102:1")
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Unexpected BGP prefix on R2"

    test_func = functools.partial(check_mpls_table, 16, "BGP")
    _, result = topotest.run_and_expect(test_func, None, count=10, wait=0.5)
    assert result is None, "Unexpected BGP label on R2"

    output = tgen.gears["r2"].vtysh_cmd("show debugging label-table")
    assert "Proto bgp: " in output, "Failed to see BGP label chunk"



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_isis_tilfa_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2020 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_isis_lfa_topo1.py:

                                       +---------+
                                       |         |
      +--------------------------------+   RT1   +-------------------------------+
      |                  +-------------+         +-------------+                 |
      |                  |             |         |             |                 |
      |                  |             +----+----+             |                 |
      |                  |                  |                  |20               |
      |                  |                  |                  |                 |
      |                  |                  |                  |                 |
 +----+----+        +----+----+        +----+----+        +----+----+       +----+----+
 |         |        |         |        |         |        |         |       |         |
 |   RT2   |   5    |   RT3   |        |   RT4   |        |   RT5   |       |   RT6   |
 |         +--------+         |        |         |        |         |       |         |
 |         |        |         |        |         |        |         |       |         |
 +----+----+        +----+----+        +----+----+        +----+----+       +----+----+
      |                  |                  |                  |                 |
      |                  |                  |15                |                 |
      |5                 |                  |                  |                 |
      |                  |             +----+----+             |                 |
      |                  |             |         |             |                 |
      |                  +-------------+   RT7   +-------------+                 |
      +--------------------------------+         +-------------------------------+
                                       |         |
                                       +---------+
"""

import os
import sys
import pytest
import json
import tempfile
from functools import partial

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# Required to instantiate the topology builder class.

pytestmark = [pytest.mark.isisd]

# Global multi-dimensional dictionary containing all expected outputs
outputs = {}


def build_topo(tgen):
    "Build function"

    #
    # Define FRR Routers
    #
    for router in ["rt1", "rt2", "rt3", "rt4", "rt5", "rt6", "rt7"]:
        tgen.add_router(router)

    #
    # Define connections
    #
    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-rt2")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-rt1")
    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-rt3")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-rt2")
    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-rt3")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-rt1")
    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-rt4")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt1")
    switch = tgen.add_switch("s5")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-rt5")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt1")
    switch = tgen.add_switch("s6")
    switch.add_link(tgen.gears["rt1"], nodeif="eth-rt6")
    switch.add_link(tgen.gears["rt6"], nodeif="eth-rt1")
    switch = tgen.add_switch("s7")
    switch.add_link(tgen.gears["rt2"], nodeif="eth-rt7")
    switch.add_link(tgen.gears["rt7"], nodeif="eth-rt2")
    switch = tgen.add_switch("s8")
    switch.add_link(tgen.gears["rt3"], nodeif="eth-rt7")
    switch.add_link(tgen.gears["rt7"], nodeif="eth-rt3")
    switch = tgen.add_switch("s9")
    switch.add_link(tgen.gears["rt4"], nodeif="eth-rt7")
    switch.add_link(tgen.gears["rt7"], nodeif="eth-rt4")
    switch = tgen.add_switch("s10")
    switch.add_link(tgen.gears["rt5"], nodeif="eth-rt7")
    switch.add_link(tgen.gears["rt7"], nodeif="eth-rt5")
    switch = tgen.add_switch("s11")
    switch.add_link(tgen.gears["rt6"], nodeif="eth-rt7")
    switch.add_link(tgen.gears["rt7"], nodeif="eth-rt6")

    #
    # Populate multi-dimensional dictionary containing all expected outputs
    #
    files = ["show_ipv6_route.ref", "show_yang_interface_isis_adjacencies.ref"]
    for rname in ["rt1", "rt2", "rt3", "rt4", "rt5", "rt6", "rt7"]:
        outputs[rname] = {}
        for step in range(1, 16 + 1):
            outputs[rname][step] = {}
            for file in files:
                if step == 1:
                    # Get snapshots relative to the expected initial network convergence
                    filename = "{}/{}/step{}/{}".format(CWD, rname, step, file)
                    outputs[rname][step][file] = open(filename).read()
                else:
                    if rname != "rt1":
                        continue
                    if file == "show_yang_interface_isis_adjacencies.ref":
                        continue

                    # Get diff relative to the previous step
                    filename = "{}/{}/step{}/{}.diff".format(CWD, rname, step, file)

                    # Create temporary files in order to apply the diff
                    f_in = tempfile.NamedTemporaryFile(mode="w")
                    f_in.write(outputs[rname][step - 1][file])
                    f_in.flush()
                    f_out = tempfile.NamedTemporaryFile(mode="r")
                    os.system(
                        "patch -s -o %s %s %s" % (f_out.name, f_in.name, filename)
                    )

                    # Store the updated snapshot and remove the temporary files
                    outputs[rname][step][file] = open(f_out.name).read()
                    f_in.close()
                    f_out.close()


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    # For all registered routers, load the zebra configuration file
    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_ISIS, os.path.join(CWD, "{}/isisd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BFD, os.path.join(CWD, "/dev/null".format(rname))
        )

    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


def router_compare_json_output(rname, command, reference, wait=0.5, count=120):
    "Compare router JSON output"

    logger.info('Comparing router "%s" "%s" output', rname, command)

    tgen = get_topogen()
    expected = json.loads(reference)

    # Run test function until we get an result. Wait at most 60 seconds.
    test_func = partial(topotest.router_json_cmp, tgen.gears[rname], command, expected)
    _, diff = topotest.run_and_expect(test_func, None, count=count, wait=wait)
    assertmsg = '"{}" JSON output mismatches the expected result'.format(rname)
    assert diff is None, assertmsg


#
# Step 1
#
# Test initial network convergence
#
def test_isis_adjacencies_step1():
    logger.info("Test (step 1): check IS-IS adjacencies")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1", "rt2", "rt3", "rt4", "rt5", "rt6", "rt7"]:
        router_compare_json_output(
            rname,
            "show yang operational-data /frr-interface:lib isisd",
            outputs[rname][1]["show_yang_interface_isis_adjacencies.ref"],
        )


def test_rib_ipv6_step1():
    logger.info("Test (step 1): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    for rname in ["rt1", "rt2", "rt3", "rt4", "rt5", "rt6", "rt7"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][1]["show_ipv6_route.ref"]
        )


#
# Step 2
#
# Action(s):
# -Disable LFA protection on all interfaces
#
# Expected changes:
# -rt1 should uninstall all backup nexthops from all routes
#
def test_rib_ipv6_step2():
    logger.info("Test (step 2): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Disabling LFA protection on all rt1 interfaces")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "no isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt3" -c "no isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt4" -c "no isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt5" -c "no isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt6" -c "no isis fast-reroute lfa"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][2]["show_ipv6_route.ref"]
        )


#
# Step 3
#
# Action(s):
# -Re-enable LFA protection on all interfaces
#
# Expected changes:
# -Revert changes from the previous step
#
def test_rib_ipv6_step3():
    logger.info("Test (step 3): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Re-enabling LFA protection on all rt1 interfaces")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt3" -c "isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt4" -c "isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt5" -c "isis fast-reroute lfa"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt6" -c "isis fast-reroute lfa"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][3]["show_ipv6_route.ref"]
        )


#
# Step 4
#
# Action(s):
# -Disable LFA load-sharing
#
# Expected changes:
# -rt1 should use at most one backup nexthop for each route
#
def test_rib_ipv6_step4():
    logger.info("Test (step 4): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Disabling LFA load-sharing on rt1")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "fast-reroute load-sharing disable"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][4]["show_ipv6_route.ref"]
        )


#
# Step 5
#
# Action(s):
# -Re-enable LFA load-sharing
#
# Expected changes:
# -Revert changes from the previous step
#
def test_rib_ipv6_step5():
    logger.info("Test (step 5): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Re-enabling LFA load-sharing on rt1")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "no fast-reroute load-sharing disable"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][5]["show_ipv6_route.ref"]
        )


#
# Step 6
#
# Action(s):
# -Limit backup computation to critical priority prefixes only
#
# Expected changes:
# -rt1 should uninstall all backup nexthops from all routes
#
def test_rib_ipv6_step6():
    logger.info("Test (step 6): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Limiting backup computation to critical priority prefixes only")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "fast-reroute priority-limit critical"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][6]["show_ipv6_route.ref"]
        )


#
# Step 7
#
# Action(s):
# -Configure a prefix priority list to classify rt7's loopback as a
#  critical-priority prefix
#
# Expected changes:
# -rt1 should install backup nexthops for rt7's loopback route.
#
def test_rib_ipv6_step7():
    logger.info("Test (step 7): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Configuring a prefix priority list")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "spf prefix-priority critical CRITICAL_DESTINATIONS"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "ipv6 access-list CRITICAL_DESTINATIONS seq 5 permit 2001:db8:1000::7/128"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][7]["show_ipv6_route.ref"]
        )


#
# Step 8
#
# Action(s):
# -Revert previous changes related to prefix priorities
#
# Expected changes:
# -Revert changes from the previous two steps
#
def test_rib_ipv6_step8():
    logger.info("Test (step 8): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Reverting previous changes related to prefix priorities")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "no ipv6 access-list CRITICAL_DESTINATIONS seq 5 permit 2001:db8:1000::7/128"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "no fast-reroute priority-limit critical"'
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "no spf prefix-priority critical CRITICAL_DESTINATIONS"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][8]["show_ipv6_route.ref"]
        )


#
# Step 9
#
# Action(s):
# -Exclude eth-rt6 from LFA computation for eth-rt2's failure
#
# Expected changes:
# -Uninstall the eth-rt2 protecting backup nexthops that go through eth-rt6
#
def test_rib_ipv6_step9():
    logger.info("Test (step 9): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Excluding eth-rt6 from LFA computation for eth-rt2's failure")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "isis fast-reroute lfa exclude interface eth-rt6"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname, "show ipv6 route isis json", outputs[rname][9]["show_ipv6_route.ref"]
        )


#
# Step 10
#
# Action(s):
# -Remove exclusion of eth-rt6 from LFA computation for eth-rt2's failure
#
# Expected changes:
# -Revert changes from the previous step
#
def test_rib_ipv6_step10():
    logger.info("Test (step 10): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info(
        "Removing exclusion of eth-rt6 from LFA computation for eth-rt2's failure"
    )
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "interface eth-rt2" -c "no isis fast-reroute lfa exclude interface eth-rt6"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][10]["show_ipv6_route.ref"],
        )


#
# Step 11
#
# Action(s):
# -Add LFA tiebreaker: prefer node protecting backup path
#
# Expected changes:
# -rt1 should prefer backup nexthops that provide node protection
#
def test_rib_ipv6_step11():
    logger.info("Test (step 11): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Adding LFA tiebreaker: prefer node protecting backup path")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "fast-reroute lfa tiebreaker node-protecting index 10"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][11]["show_ipv6_route.ref"],
        )


#
# Step 12
#
# Action(s):
# -Add LFA tiebreaker: prefer backup path via downstream node
#
# Expected changes:
# -rt1 should prefer backup nexthops that satisfy the downstream condition
#
def test_rib_ipv6_step12():
    logger.info("Test (step 12): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Adding LFA tiebreaker: prefer backup path via downstream node")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "fast-reroute lfa tiebreaker downstream index 20"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][12]["show_ipv6_route.ref"],
        )


#
# Step 13
#
# Action(s):
# -Add LFA tiebreaker: prefer backup path with lowest total metric
#
# Expected changes:
# -rt1 should prefer backup nexthops that have the best metric
#
def test_rib_ipv6_step13():
    logger.info("Test (step 13): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Adding LFA tiebreaker: prefer backup path with lowest total metric")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "fast-reroute lfa tiebreaker lowest-backup-metric index 30"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][13]["show_ipv6_route.ref"],
        )


#
# Step 14
#
# Action(s):
# - Setting spf-delay-ietf init-delay of 15s
#
# Expected changes:
# - No routing table change
# - At the end of test, SPF reacts to a failure in 15s
#
def test_rib_ipv6_step14():
    logger.info("Test (step 14): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Setting spf-delay-ietf init-delay of 15s")
    tgen.net["rt1"].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "spf-delay-ietf init-delay 15000 short-delay 0 long-delay 0 holddown 0 time-to-learn 0"'
    )

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][14]["show_ipv6_route.ref"],
        )


#
# Step 15
#
# Action(s):
# - shut the eth-rt2 interface on rt1
#
# Expected changes:
# - Route switchover of routes via eth-rt2
#
def test_rib_ipv6_step15():
    logger.info("Test (step 15): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Shut the interface to rt2 from the switch side and check fast-reroute")
    tgen.net.cmd_raises("ip link set %s down" % tgen.net["s1"].intfs[0])

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][15]["show_ipv6_route.ref"],
            count=10,
            wait=0.5,
        )


#
# Step 16
#
# Action(s): wait for the convergence and SPF computation on rt1
#
# Expected changes:
# - convergence of IPv6 RIB
#
def test_rib_ipv6_step16():
    logger.info("Test (step 16): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Check SPF convergence")

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][16]["show_ipv6_route.ref"],
        )


#
# Step 17
#
# Action(s):
# - Unshut the interface to rt2 from the switch sid
#
# Expected changes:
# - The routing table converges
#
def test_rib_ipv6_step17():
    logger.info("Test (step 17): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    rname = "rt1"

    logger.info("Unsetting spf-delay-ietf init-delay of 15s")
    tgen.net[rname].cmd('vtysh -c "conf t" -c "router isis 1" -c "no spf-delay-ietf"')

    logger.info(
        "Unshut the interface to rt2 from the switch side and check fast-reroute"
    )
    tgen.net.cmd_raises("ip link set %s up" % tgen.net["s1"].intfs[0])

    logger.info("Setting spf-delay-ietf init-delay of 15s")
    tgen.net[rname].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "spf-delay-ietf init-delay 15000 short-delay 0 long-delay 0 holddown 0 time-to-learn 0"'
    )

    router_compare_json_output(
        rname,
        "show ipv6 route isis json",
        outputs[rname][14]["show_ipv6_route.ref"],
    )


#
# Step 18
#
# Action(s):
# - drop traffic between rt1 and rt2 by shutting down the bridge between
#   the routers. Interfaces on rt1 and rt2 stay up.
#
#
# Expected changes:
# - Route switchover of routes via eth-rt2
#
def test_rib_ipv6_step18():
    def _rt2_neigh_down(router):
        output = json.loads(router.vtysh_cmd("show isis neighbor rt2 json"))

        """
        Previous output was:
        {
          "areas":[
            {
              "area":"1",
              "circuits":[
                {
                  "circuit":0,
                  "adj":"rt2",
                  "interface":{
                    "name":"eth-rt2",
                    "state":"Up",
                    "adj-flaps":1,
                    "last-ago":"21s",
                    "circuit-type":"L1",
                    "speaks":"IPv6",
                    "topologies":{
                      "topo-0":"ipv6-unicast"
                    },
                    "snpa":"2020.2020.2020",
                    "area-address":{
                      "isonet":"49.0000"
                    },
                    "ipv6-link-local":{
                      "ipv6":"fe80::ac19:a8ff:fee5:f48f"
                    },
                    "adj-sid":{
                    }
                  },
                  "level":1,
                  "expires-in":"2s"
                },
                {
                  "circuit":0
                },
                {
                  "circuit":0
                },
                {
                  "circuit":0
                },
                {
                  "circuit":0
                },
                {
                  "circuit":0
                }
              ]
            }
          ]
        """

        expected = {
            "areas": [
                {
                    "area": "1",
                    "circuits": [
                        {"circuit": 0},
                        {"circuit": 0},
                        {"circuit": 0},
                        {"circuit": 0},
                        {"circuit": 0},
                        {"circuit": 0},
                    ],
                }
            ]
        }

        return topotest.json_cmp(output, expected, exact=True)

    logger.info("Test (step 18): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Drop traffic between rt1 and rt2")
    tgen.net.cmd_raises("ip link set s1 down")

    rname = "rt1"
    router = tgen.gears[rname]
    test_func = partial(_rt2_neigh_down, router)
    _, result = topotest.run_and_expect(test_func, None, count=200, wait=0.05)
    assert result is None, 'rt2 neighbor is still present on "{}"'.format(router)

    router_compare_json_output(
        rname,
        "show ipv6 route isis json",
        outputs[rname][15]["show_ipv6_route.ref"],
        count=10,
        wait=0.5,
    )


#
# Step 19
#
# Action(s): wait for the convergence and SPF computation on rt1
#
# Expected changes:
# - convergence of IPv6 RIB
#
def test_rib_ipv6_step19():
    logger.info("Test (step 19): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Check SPF convergence")

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][16]["show_ipv6_route.ref"],
        )


#
# Step 20
#
# Action(s):
# - Unshut the switch from rt1 to rt2
#
# Expected changes:
# - The routing table goes back to the nominal state
#
def test_rib_ipv6_step20():
    logger.info("Test (step 20): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    rname = "rt1"

    logger.info("Unsetting spf-delay-ietf init-delay of 15s")
    tgen.net[rname].cmd('vtysh -c "conf t" -c "router isis 1" -c "no spf-delay-ietf"')

    logger.info(
        "Unshut the interface to rt2 from the switch side and check fast-reroute"
    )
    tgen.net.cmd_raises("ip link set s1 up")

    logger.info("Setting spf-delay-ietf init-delay of 15s")
    tgen.net[rname].cmd(
        'vtysh -c "conf t" -c "router isis 1" -c "spf-delay-ietf init-delay 15000 short-delay 0 long-delay 0 holddown 0 time-to-learn 0"'
    )

    router_compare_json_output(
        rname,
        "show ipv6 route isis json",
        outputs[rname][14]["show_ipv6_route.ref"],
    )


#
# Step 21
#
# Action(s):
# - clear the rt2 ISIS neighbor on rt1
#
# Expected changes:
# - Route switchover of routes via eth-rt2
#
def test_rib_ipv6_step21():
    logger.info("Test (step 21): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    rname = "rt1"

    logger.info("Clear the rt2 ISIS neighbor on rt1 and check fast-reroute")
    tgen.gears[rname].vtysh_cmd("clear isis neighbor rt2")

    router_compare_json_output(
        rname,
        "show ipv6 route isis json",
        outputs[rname][15]["show_ipv6_route.ref"],
        count=10,
        wait=0.5,
    )


#
# Step 22
#
# Action(s): wait for the convergence and SPF computation on rt1
#
# Expected changes:
# - convergence of IPv6 RIB
#
def test_rib_ipv6_step22():
    logger.info("Test (step 22): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Check SPF convergence")

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][16]["show_ipv6_route.ref"],
        )


#
# Step 23
#
# Action(s):
# - Setting BFD
#
# Expected changes:
# - No routing table change
# - BFD comes up
#
def test_rib_ipv6_step23():
    logger.info("Test (step 23): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Setup BFD on rt1 and rt2")
    for rname in ["rt1", "rt2"]:
        conf_file = os.path.join(CWD, "{}/bfdd.conf".format(rname))
        tgen.net[rname].cmd("vtysh -f {}".format(conf_file))

    logger.info("Set ISIS BFD")
    tgen.net["rt1"].cmd('vtysh -c "conf t" -c "int eth-rt2" -c "isis bfd"')
    tgen.net["rt2"].cmd('vtysh -c "conf t" -c "int eth-rt1" -c "isis bfd"')

    rname = "rt1"
    expect = '[{"multihop":false,"interface":"eth-rt2","status":"up"}]'
    router_compare_json_output(rname, "show bfd peers json", expect)

    router_compare_json_output(
        rname,
        "show ipv6 route isis json",
        outputs[rname][14]["show_ipv6_route.ref"],
    )


#
# Step 24
#
# Action(s):
# - drop traffic between rt1 and rt2 by shutting down the bridge between
#   the routers. Interfaces on rt1 and rt2 stay up.
#
# Expected changes:
# - BFD comes down before IS-IS
# - Route switchover of routes via eth-rt2
#
def test_rib_ipv6_step24():
    def _bfd_down(router):
        output = json.loads(router.vtysh_cmd("show bfd peers json"))
        expected = []
        return topotest.json_cmp(output, expected, exact=True)

    logger.info("Test (step 24): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Shut the interface to rt2 from the switch side and check fast-reroute")
    tgen.net.cmd_raises("ip link set s1 down")

    rname = "rt1"
    router = tgen.gears[rname]
    test_func = partial(_bfd_down, router)
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=0.3)
    assert result is None, 'BFD session is still up on "{}"'.format(router)

    router_compare_json_output(
        rname,
        "show ipv6 route isis json",
        outputs[rname][15]["show_ipv6_route.ref"],
        count=10,
    )


#
# Step 25
#
# Action(s): wait for the convergence and SPF computation on rt1
#
# Expected changes:
# - convergence of IPv6 RIB
#
def test_rib_ipv6_step25():
    logger.info("Test (step 25): verify IPv6 RIB")
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Check SPF convergence")

    for rname in ["rt1"]:
        router_compare_json_output(
            rname,
            "show ipv6 route isis json",
            outputs[rname][16]["show_ipv6_route.ref"],
        )


# Memory leak test template
def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
